import struct
from io import BytesIO

from luabins.constants import *


def _read_bytes_or_raise(stream: BytesIO, num_bytes: int) -> bytes:
    ret_bytes = stream.read(num_bytes)

    if len(ret_bytes) != num_bytes:
        raise Exception(f"Tried to get {num_bytes} but got {len(ret_bytes)}")

    return ret_bytes


def _read_short_short_int(stream: BytesIO) -> int:
    return int.from_bytes(_read_bytes_or_raise(stream, LUA_SHORT_SHORT_INT_LENGTH), LUA_BYTEORDER)


def _read_int(stream: BytesIO) -> int:
    return int.from_bytes(_read_bytes_or_raise(stream, LUA_INT_LENGTH), LUA_BYTEORDER)


def _read_number(stream: BytesIO) -> float:
    return struct.unpack("<d", _read_bytes_or_raise(stream, LUA_NUMBER_LENGTH))[0]


def _read_string(stream: BytesIO) -> str:
    length = _read_int(stream)

    string_bytes = _read_bytes_or_raise(stream, length)

    if not string_bytes:
        return ""

    return string_bytes.decode(LUA_STR_ENCODING)


def _build_short_short_int(value: int) -> bytes:
    return value.to_bytes(LUA_SHORT_SHORT_INT_LENGTH, LUA_BYTEORDER)


def _build_int(value: int) -> bytes:
    return value.to_bytes(LUA_INT_LENGTH, LUA_BYTEORDER)


def _build_number(value: float) -> bytes:
    return struct.pack("<d", value)


def _build_string(value: str) -> bytes:
    str_bytes = value.encode(LUA_STR_ENCODING)
    len_bytes = len(str_bytes).to_bytes(LUA_INT_LENGTH, LUA_BYTEORDER)

    return len_bytes + str_bytes



import math
from io import BytesIO
from typing import Any, List, Dict

from luabins.byte_utils import _read_int, _read_short_short_int, _read_number, _read_string, _build_number, \
    _build_string
from luabins.constants import *
from luabins.lua_table_key import LuaTableKey


def _read_table(stream: BytesIO) -> Dict[Any, Any]:
    table = {}

    array_size = _read_int(stream)
    hash_size = _read_int(stream)
    total_size = array_size + hash_size

    if array_size < 0 or hash_size < 0:
        raise Exception("Invalid table sizes")

    for i in range(total_size):
        key = _load_value(stream)
        if isinstance(key, dict):
            key = LuaTableKey(key)
        if key is None:
            raise Exception("Key in a table cannot be none")
        elif isinstance(key, float) and math.isnan(key):
            raise Exception("Key may not be NaN")

        table[key] = _load_value(stream)

    return table


def _build_table(table: Dict[Any, Any]) -> bytes:
    # This isn't TOTALLY accurate, but it's also not important that it's perfect
    array_size = len([key for key in table.keys() if isinstance(key, int)])
    hash_size = len(table.keys()) - array_size

    result = array_size.to_bytes(LUA_INT_LENGTH, LUA_BYTEORDER) + hash_size.to_bytes(LUA_INT_LENGTH, LUA_BYTEORDER)

    for (key, value) in table.items():
        if isinstance(key, LuaTableKey):
            key = key.inner
        result += _save_value(key)
        result += _save_value(value)

    return result


def _load_value(stream: BytesIO) -> Any:
    value_type = _read_short_short_int(stream)

    if value_type == LUABINS_NIL:
        return None
    elif value_type == LUABINS_FALSE:
        return False
    elif value_type == LUABINS_TRUE:
        return True
    elif value_type == LUABINS_NUMBER:
        return _read_number(stream)
    elif value_type == LUABINS_STRING:
        return _read_string(stream)
    elif value_type == LUABINS_TABLE:
        return _read_table(stream)
    else:
        raise Exception(f"Unknown type {value_type}")


def _save_value(value: Any) -> bytes:
    if value is None:
        return LUABINS_NIL.to_bytes(1, LUA_BYTEORDER)
    if value is False:
        return LUABINS_FALSE.to_bytes(1, LUA_BYTEORDER)
    if value is True:
        return LUABINS_TRUE.to_bytes(1, LUA_BYTEORDER)
    if isinstance(value, (int, float)):
        return LUABINS_NUMBER.to_bytes(1, LUA_BYTEORDER) + _build_number(float(value))
    if isinstance(value, str):
        return LUABINS_STRING.to_bytes(1, LUA_BYTEORDER) + _build_string(value)
    if isinstance(value, dict):
        return LUABINS_TABLE.to_bytes(1, LUA_BYTEORDER) + _build_table(value)
    if isinstance(value, list):
        as_dict = {index + 1: value_at_index for (index, value_at_index) in enumerate(value)}
        return LUABINS_TABLE.to_bytes(1, LUA_BYTEORDER) + _build_table(as_dict)
    else:
        raise Exception(f"Unknown type {type(value)}")


def decode_luabins(stream: BytesIO, respect_max_items: bool = True) -> List[Any]:
    num_items = _read_short_short_int(stream)

    if num_items > 250 and respect_max_items:
        raise Exception("Max items in a serialized blob for luabin is 250")

    values = []

    for i in range(num_items):
        values.append(_load_value(stream))

    if len(stream.read(1)) != 0:
        raise Exception(f"Read {num_items} values, but we still have more data in the stream! Data corrupt?")

    return values


def encode_luabins(values: List[Any]) -> bytes:
    output = len(values).to_bytes(1, LUA_BYTEORDER)

    for value in values:
        output += _save_value(value)

    return output



from io import BytesIO
from typing import List, Any

import pytest

from luabins import encode_luabins, decode_luabins


def assert_save_load(value: List[Any], comparison_value=None):
    comparison_value = comparison_value or value
    assert decode_luabins(BytesIO(encode_luabins(value))) == comparison_value


def test_nil():
    assert_save_load([None])


def test_true():
    assert_save_load([True])


def test_false():
    assert_save_load([False])


def test_numbers():
    assert_save_load([0])
    assert_save_load([1])
    assert_save_load([-1])
    assert_save_load([0.0])
    assert_save_load([1.0])
    assert_save_load([-1.0])
    assert_save_load([0.5])
    assert_save_load([-0.5])


def test_string():
    assert_save_load([""])
    assert_save_load(["test"])


def test_complex_string():
    assert_save_load(["sauté"])


def test_table():
    assert_save_load([{}])
    assert_save_load([{"test": "test"}])
    assert_save_load([{1.0: "test"}])
    assert_save_load([{1.0: 0}])
    assert_save_load([{0: "test"}])
    assert_save_load([{0: 0.0}])
    assert_save_load([{True: True}])
    assert_save_load([{False: False}])
    assert_save_load([{"test": {"test": "test"}}])
    assert_save_load([[]], [{}])
    assert_save_load([["test"]], [{1.0: "test"}])


def test_errors():
    data = [i for i in range(252)]

    with pytest.raises(Exception):
        assert_save_load(data)

    with pytest.raises(Exception):
        assert_save_load([{None: "test"}])

    with pytest.raises(Exception):
        assert_save_load([float("bad")])



LUABINS_NIL = 0x2D
LUABINS_FALSE = 0x30
LUABINS_TRUE = 0x31
LUABINS_NUMBER = 0x4E
LUABINS_STRING = 0x53
LUABINS_TABLE = 0x54

LUA_SHORT_SHORT_INT_LENGTH = 1
LUA_INT_LENGTH = 4
LUA_SIZE_T_LENGTH = 4
LUA_NUMBER_LENGTH = 8

LUA_BYTEORDER = "little"
LUA_STR_ENCODING = "utf-8"



import json


class LuaTableKey:
    def __init__(self, inner):
        self.inner = inner

    def __hash__(self):
        return json.dumps(self.inner).__hash__()

    def __eq__(self, other):
        return other.__hash__() == self.__hash__()



import os
from io import BytesIO

from luabins import decode_luabins, encode_luabins


def test_debug():
    path = os.path.join(os.path.dirname(__file__), "debug.bin")
    with open(path, "rb") as f:
        orig = f.read()
        decode1 = decode_luabins(BytesIO(orig))
        re_encoded = encode_luabins(decode1)
        decode2 = decode_luabins(BytesIO(re_encoded))
        assert decode2 == decode1



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='luabins_py',
    version='1.4.0',
    packages=['luabins'],
    url='https://github.com/zsennenga/luabins_py',
    license='MIT License',
    author='Zach Ennenga',
    author_email='astrozees@gmail.com',
    description='A library to encode/decode luabins serialized data in python',
    long_description=long_description,
    long_description_content_type='text/markdown',
)



import os

from luabins import decode_luabins, encode_luabins


def open_and_assert(filename, expected):
    path = os.path.join(os.path.dirname(__file__), filename)
    with open(path, "rb") as f:
        assert decode_luabins(f) == expected

    path = os.path.join(os.path.dirname(__file__), "output_" + filename)
    with open(path, "wb") as f:
        f.write(encode_luabins(expected))


def test_integration():
    open_and_assert("basic.bin", [1, "two", {1.0: "three", 2.0: 4}])
    open_and_assert("basic2.bin", [1, "two", {1.0: "three", 2.0: 4.0}])
    open_and_assert("basic3.bin", [1, "two", {1.0: "three", 2.0: {1.0: "four", 2.0: 4.0}}])
    open_and_assert("basic4.bin", [None, True, False])
    open_and_assert("basic5.bin", [{"test": "test"}])




